from typing import Any, Iterable

import matplotlib.pyplot as plt
from matplotlib import animation
from mpl_toolkits.mplot3d import axes3d
import numpy as np
import numpy.linalg as lin
from uncertainties import ufloat, ufloat_fromstr, unumpy, UFloat

from hikari.dataframes import BaseFrame, CifFrame, UBaseFrame
from hikari.utility import make_abspath, cfloat, det3x3, chemical_elements, \
    rotation_around


def calculate_similarity_indices(cif1_path: str,
                                 cif2_path: str = None,
                                 cif1_block: str = None,
                                 cif2_block: str = None,
                                 output_path: str = None,
                                 normalize: bool = False,
                                 uncertainties: bool = True) -> None:
    """
    Compare Anisotropic Displacement Parameters of all atoms in two cif blocks
    and return a Similarity Index (SI) for these pairs of atoms that exist
    and are defined anisotropic (via "_atom_site_aniso_U_**") in both files.

    This script requires two independent cif blocks to run,
    but the location of both blocks can be specified in multiple ways:

    - If only `cif1_path` is given, the first and second data block from
      cif1 file will be used in SI evaluation;
    - If `cif1_path` and `cif2_path` are given, the first data block from
      each cif file will be used instead;
    - If `cif1_path`, `cif2_path`, and `cif1_block` are given,
      `cif1_block` from each `cif1_path` and `cif2_path` cif files will be used.
    - If `cif1_path`, `cif2_path`, `cif1_block`, and `cif2_block` are given,
      `cif1_block` in `cif1_path` and `cif2_block` in `cif2_path` will be used.

    For a single cif file 'glycine.cif' with two data blocks named '100K' and
    'RT', the three following commands will have the same effect:

    :Example:

    >>> calculate_similarity_indices('glycine.cif')
    >>> calculate_similarity_indices('glycine.cif', 'glycine.cif')
    >>> calculate_similarity_indices('glycine.cif', 'glycine.cif', '100K', 'RT')

    For two cif files 'X-rays.cif' and 'neutrons.cif' with data block named 'RT'
    and '100K' (one each), the two following commands will have the same effect:

    :Example:

    >>> calculate_similarity_indices('X-rays.cif', 'neutrons.cif')
    >>> calculate_similarity_indices('X-rays.cif', 'neutrons.cif', 'RT', '100K')

    The behaviour of script can be further altered via other parameters.
    If `output_file` is set True, the results will be written there instead of
    console. If `uncertainties` is set True, standard deviations of all ADPs
    as well as the unit cell parameters from 1st cif file will be assumed
    uncorrelated and used to estimate the uncertainty of every SI determination.

    For more information about Similarity Index (SI) itself, please consult
    Whitten and Spackman, Acta Cryst B **62**, 875 (2006)
    https://doi.org/10.1107/S0108768106020787.

    :param cif1_path: Absolute or relative path to the first cif file.
    :param cif2_path: Absolute or relative path to the second cif file.
        If not specified, it is assumed equal to `cif1_path`.
    :param cif1_block: Name of the first data block used in SI determination.
        It points to the data block inside the file specified by `cif1_path`.
        If not specified, the first block found in said file will be used.
    :param cif2_block: Name of the second data block used in SI determination.
        It points to the data block inside the file specified by `cif2_path`.
        If not specified, the first unused block in said file will be used.
    :param output_path: Path where the output of the program should be written.
    :param uncertainties: If True, propagate the standard deviations of
        individual ADPs' and cif1's unit cell to estimate SI's uncertainties.
    :param normalize: If True, equalize the volume of displacement ellipsoids
        by normalizing the determinants of ADP matrices expressed in cartesian
        coordinates. As a result, SI is a function of displacement "shape" only.
    """

    u_type = ufloat_fromstr if uncertainties else cfloat

    def initialise_output():
        if output_path is None:
            f_ = None
        else:
            f_ = open(make_abspath(output_path), 'w+')
        return f_
    f = initialise_output()

    def interpret_paths():
        cif1p = make_abspath(cif1_path)
        cif2p = cif1p if cif2_path is None else make_abspath(cif2_path)
        cif1b = 1 if cif1_block is None else cif1_block
        cif2b = 2 if cif1b == 1 and cif1p is cif2p \
            else 1 if cif2_block is None else cif2_block
        return cif1p, cif2p, cif1b, cif2b
    cif1_path, cif2_path, cif1_block, cif2_block = interpret_paths()

    print(f"# Hikari will calculate similarity indices for atoms "
          f"in the following blocks:", file=f)
    print(f"# - Block {cif1_block} of file {cif1_path}", file=f)
    print(f"# - Block {cif2_block} of file {cif2_path}", file=f)
    print(f"# with the following settings:", file=f)
    print(f"# - normalize={normalize}", file=f)
    print(f"# - uncertainties={uncertainties}", file=f)

    def read_cif_block(path, block):
        c = CifFrame()
        c.read(path)
        block = list(c.keys())[block - 1] if isinstance(block, int) else block
        return c[block]
    cif_block_1 = read_cif_block(cif1_path, cif1_block)
    cif_block_2 = read_cif_block(cif2_path, cif2_block)

    def find_common_labels(block1, block2):
        label_list1 = block1['_atom_site_aniso_label']
        label_list2 = block2['_atom_site_aniso_label']
        return [label for label in label_list1 if label in label_list2]
    common_labels = find_common_labels(cif_block_1, cif_block_2)
    print(f"# Number of matching atoms found: {len(common_labels)}", file=f)

    def make_adp_fractional_arrays(block):
        labels = block['_atom_site_aniso_label']
        u11s = block.get_as_type(key='_atom_site_aniso_U_11', typ=u_type)
        u22s = block.get_as_type(key='_atom_site_aniso_U_22', typ=u_type)
        u33s = block.get_as_type(key='_atom_site_aniso_U_33', typ=u_type)
        u12s = block.get_as_type(key='_atom_site_aniso_U_12', typ=u_type)
        u13s = block.get_as_type(key='_atom_site_aniso_U_13', typ=u_type)
        u23s = block.get_as_type(key='_atom_site_aniso_U_23', typ=u_type)
        return {k: np.array([[u11, u12, u13], [u12, u22, u23], [u13, u23, u33]])
                for k, u11, u22, u33, u12, u13, u23
                in zip(labels, u11s, u22s, u33s, u12s, u13s, u23s)}
    adp_frac_dict_1 = make_adp_fractional_arrays(cif_block_1)
    adp_frac_dict_2 = make_adp_fractional_arrays(cif_block_2)

    base_frame1 = UBaseFrame() if uncertainties else BaseFrame()
    base_frame2 = UBaseFrame() if uncertainties else BaseFrame()
    base_frame1.fill_from_cif_block(cif_block_1)
    base_frame2.fill_from_cif_block(cif_block_2)

    def calculate_similarity_index(adp_frac_1, adp_frac_2) -> float or UFloat:
        def adp_frac2cart(adp_frac, base_frame):
            n = np.diag([base_frame.a_r, base_frame.b_r, base_frame.c_r])
            return base_frame.A_d.T @ n @ adp_frac @ n @ base_frame.A_d
        try:
            adp_cart_1 = adp_frac2cart(adp_frac_1, base_frame1)
            adp_cart_2 = adp_frac2cart(adp_frac_2, base_frame2)
            if normalize:
                adp_cart_1 /= det3x3(adp_cart_1) ** (1 / 3)
                adp_cart_2 /= det3x3(adp_cart_2) ** (1 / 3)
            if uncertainties:
                # `unumpy.matrix` and `unumpy.ulinalg.pinv` are deprecated and
                # will break soon, but uncertainties is ~no longer supported...
                adp_inv_1 = unumpy.matrix(adp_cart_1).I
                adp_inv_2 = unumpy.matrix(adp_cart_2).I
            else:
                adp_inv_1 = np.linalg.inv(adp_cart_1)
                adp_inv_2 = np.linalg.inv(adp_cart_2)
            r12_num = 2 ** (3 / 2) * det3x3(adp_inv_1 @ adp_inv_2) ** (1 / 4)
            r12_den = det3x3(adp_inv_1 + adp_inv_2) ** (1 / 2)
        except ValueError:  # if matrix cannot be inverted, return S = 1.0
            r12_num = ufloat(0.0, 0) if uncertainties else 0.0
            r12_den = ufloat(1.0, 0) if uncertainties else 1.0
        return 100 * (1 - r12_num / r12_den)

    def si_string(si: float or UFloat) -> str:
        return f'{si:8.4f}' if isinstance(si, float) \
            else f'{si.n:8.4f} +/- {si.s:8.4f}'

    def si_header():
        return '#  label     SI.n +/-     SI.s' \
            if uncertainties else '#  label       SI'
    print(si_header(), file=f)

    def calculate_and_print_similarity_indices():
        sis: dict[Any, float or UFloat] = {}
        for k in common_labels:
            sis[k] = calculate_similarity_index(adp_frac_dict_1[k],
                                                adp_frac_dict_2[k])
            print(f'{k:>8} {si_string(sis[k])}', file=f)
        sis_all = [v for k, v in sis.items()]
        sis_h = [v for k, v in sis.items()
                 if k[0] == 'H' and k[:2] not in chemical_elements]
        if sis_all:
            avg_si_all = sum(sis_all) / len(sis_all)
            print(f'# avg(*) {si_string(avg_si_all)}', file=f)
        else:
            print(f'# No atoms with matching names found', file=f)
        if sis_h:
            avg_si_h = sum(sis_h) / len(sis_h)
            print(f'# avg(H) {si_string(avg_si_h)}', file=f)
    calculate_and_print_similarity_indices()

    def terminate_output():
        if output_path is not None:
            f.close()
    terminate_output()


def animate_similarity_index(u_diag: Iterable,
                             transformations: list[np.ndarray],
                             output_path: str) -> None:
    """
    Make a gif presenting the change of similarity index against some arbitrary
    series of `transformations`. Initial displacement matrix must be diagonal,
    values given in `u_diag`, evaluated against a regular unit cell with a = 1.

    :param u_diag: triplet of u matrix diagonal elements
    :param transformations: list of transformations to be plotted every 0.1 sec
    :param output_path: path where resulting gif file will be saved
    """
    max_radius = 1.0

    fig = plt.figure(figsize=(12, 12))  # Square figure

    # Coefficients in a0 x**2 + a1 y**2 + a2 z**2 = 1 & corresponding radii
    rx, ry, rz = np.sqrt(u_diag)
    u = np.diag(u_diag)

    # Set of all spherical angles:
    v = np.linspace(0, 2 * np.pi, 100)
    w = np.linspace(0, np.pi, 100)

    # Cartesian coordinates that correspond to the spherical angles:
    # (this is the equation of an ellipsoid):
    x = rx * np.outer(np.cos(v), np.sin(w))
    y = ry * np.outer(np.sin(v), np.sin(w))
    z = rz * np.outer(np.ones_like(v), np.cos(w))

    # set animation settings
    fps = 10
    steps = len(transformations)

    def calculate_similarity_index(adp_frac_1, adp_frac_2) -> float or UFloat:
        base_frame = BaseFrame()

        def adp_frac2cart(adp_frac):
            n = np.diag([base_frame.a_r, base_frame.b_r, base_frame.c_r])
            return base_frame.A_d.T @ n @ adp_frac @ n @ base_frame.A_d
        adp_cart_1 = adp_frac2cart(adp_frac_1)
        adp_cart_2 = adp_frac2cart(adp_frac_2)
        adp_inv_1 = np.linalg.inv(adp_cart_1)
        adp_inv_2 = np.linalg.inv(adp_cart_2)
        r12_num = 2 ** (3 / 2) * det3x3(adp_inv_1 @ adp_inv_2) ** (1 / 4)
        r12_den = det3x3(adp_inv_1 + adp_inv_2) ** (1 / 2)
        return 100 * (1 - r12_num / r12_den)

    def get_xyz(step: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get positions of x, y, z nodes after `step` steps of animation"""
        trans_matrix = transformations[step]
        x_list = x.reshape(10_000)
        y_list = y.reshape(10_000)
        z_list = z.reshape(10_000)
        xyz = np.vstack([x_list, y_list, z_list]).T
        xyz_trans = xyz @ trans_matrix
        x_trans = xyz_trans[:, 0].reshape(100, 100)
        y_trans = xyz_trans[:, 1].reshape(100, 100)
        z_trans = xyz_trans[:, 2].reshape(100, 100)
        return x_trans, y_trans, z_trans

    def get_u(step: int) -> np.ndarray:
        """Get transformed u matrix after `step` steps of animation"""
        trans_matrix = transformations[step]
        return trans_matrix.T @ u @ trans_matrix

    # prepare list of similarity indices
    si = [calculate_similarity_index(u, get_u(i)) for i in range(steps)]
    si.append(si[0])

    def get_frame(step):
        """Clear previous axes and produce new at `step` step of animation"""
        for a in fig.axes:
            a.clear()
        x_, y_, z_ = get_xyz(step)
        ax1 = fig.add_subplot(221, projection='3d')
        ax2 = fig.add_subplot(222)
        ax3 = fig.add_subplot(223, projection='3d')
        ax4 = fig.add_subplot(224)
        ax1.plot_surface(x, y, z, rstride=4, cstride=4, color='r')
        ax2.plot(range(steps+1), si, 'b')
        ax2.set_xlabel(step)
        ax2.set_ylabel('similarity index')
        ax2.axvline(x=step, color='b')
        ax2.plot(step, si[step], 'bo')
        ax2.set_xlim([0, steps])
        ax2.set_ylim([0, 10])
        ax3.plot_surface(x_, y_, z_, rstride=4, cstride=4, color='b')
        u_ = get_u(step)
        ax4_font = {'family': 'serif', 'weight': 'normal', 'size': 16}
        ax4.clear()
        ax4.axis('off')
        ax4.text(0.2, 0.5, s='U = [', fontdict=ax4_font, ha='right')
        ax4.text(0.4, 0.7, s=f'{u_[0, 0]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.6, 0.7, s=f'{u_[1, 0]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.8, 0.7, s=f'{u_[2, 0]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.4, 0.5, s=f'{u_[0, 1]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.6, 0.5, s=f'{u_[1, 1]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.8, 0.5, s=f'{u_[2, 1]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.4, 0.3, s=f'{u_[0, 2]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.6, 0.3, s=f'{u_[1, 2]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.8, 0.3, s=f'{u_[2, 2]:6.3f}', fontdict=ax4_font, ha='right')
        ax4.text(0.9, 0.5, s=']', fontdict=ax4_font, ha='right')
        for axis in 'xyz':
            getattr(ax1, f'set_{axis}lim')((-max_radius, max_radius))
            getattr(ax3, f'set_{axis}lim')((-max_radius, max_radius))

    fig.tight_layout()
    fig.subplots_adjust(left=0.02, bottom=0.02, right=0.98, top=0.98,
                        wspace=None, hspace=None)
    ani = animation.FuncAnimation(
        fig=fig,
        func=get_frame,
        frames=steps,
        interval=1000/fps,
    )
    ani_path = make_abspath(output_path)
    ani.save(ani_path, writer='imagemagick', fps=fps)


if __name__ == '__main__':
    # calculate_similarity_indices('~/_/si/1.cif',
    #                              '~/_/si/2.cif',
    #                              output_path='~/_/si/output.txt')
    t1 = [rotation_around(np.array([0, 0, 1]), by=np.deg2rad(10 * i))
          for i in range(36)]
    t2 = [np.eye(3) * np.sqrt((1.5 - np.cos(np.deg2rad(10 * i)) / 2))
          for i in range(36)]
    t3 = [np.array([(1, 0, 0), (0, 1, 0), (0, 0, np.sqrt(2 ** np.sin(np.deg2rad(10 * i))))])
          for i in range(36)]
    animate_similarity_index(u_diag=(2, 1, 1),
                             transformations=t1,
                             output_path='~/_/si/S_animation_211_zrot.gif')



import io
import json

import pandas as pd
from importlib.resources import open_text, open_binary


def _load_indexed_csv(resource_name: str) -> pd.DataFrame:
    with open_text(__name__, resource_name, encoding='utf-8') as f:
        s = io.StringIO(f.read())
    return pd.read_csv(s, comment='#', index_col=0)


def _load_indexed_wsv(resource_name: str) -> pd.DataFrame:
    with open_text(__name__, resource_name, encoding='utf-8') as f:
        s = io.StringIO(f.read())
    return pd.read_csv(s, comment='#', index_col=0, sep=r'\s+')


def _load_json(resource_name: str) -> dict:
    with open_text(__name__, resource_name, encoding='utf-8') as f:
        return json.load(f)


def _load_text(resource_name: str) -> str:
    with open_text(__name__, resource_name, encoding='utf-8') as f:
        return f.read()


gnuplot_angular_heatmap_template = _load_text('gnuplot_angular_heatmap_template.gnu')
point_groups_json = _load_text('point_groups.json')
space_groups_json = _load_text('space_groups.json')
hkl_formats = _load_json('hkl_formats_defined.json')
hkl_aliases = _load_json('hkl_formats_aliases.json')
hkl_mercury_style = _load_text('hkl.msd')
characteristic_radiation = _load_json('characteristic_radiation.json')
cif_core_dict = _load_text('cif_core_2.4.5.dic')
Xray_atomic_form_factors = _load_indexed_csv('Xray_atomic_form_factors.csv')
point_groups_dataframe = _load_indexed_wsv('point_groups.wsv')
space_groups_dataframe = _load_indexed_wsv('space_groups.wsv')



"""
This file contains tools to work with dictionaries used in the package.
"""


def dict_union(*dicts: dict) -> dict:
    """Return a union of dicts; if conflicts, later dicts take precedence."""
    return dict(i for d in dicts for i in d.items())



"""
The module containing relatively simple functions,
which are not necessarily strictly connected
to the actual contents of the whole library,
but do significantly improve clarity of the code present in other modules.
"""

from .certain_float import cfloat
from .chem_tools import chemical_elements, split_atom_label
from .dict_tools import dict_union
from .interval import Interval
from .math_tools import angle2rad, cart2sph, sph2cart, det3x3, \
    fibonacci_sphere, rotation_from, rotation_around, weighted_quantile
from .list_tools import cubespace, find_best, rescale_list_to_range,\
    rescale_list_to_other
from .os_tools import make_abspath
from .palettes import gnuplot_map_palette, mpl_map_palette
from .artists import artist_factory
from .numpy_tools import str2array
from .singleton import Singleton



import numpy as np
from matplotlib import pyplot, cm
from scipy.optimize import minimize
from scipy.special import erfinv
from scipy.stats import norm

from hikari.dataframes import HklFrame, ResFrame
from hikari.symmetry import SG
from hikari.utility import make_abspath


def baycon_plot(x_key='ze', y_key='si',
                a=10.0, b=10.0, c=10.0, al=90.0, be=90.0, ga=90.0,
                input_path='shelx.fcf',
                input_format='shelx_fcf',
                input_wavelength='MoKa',
                output_path='baycon.png'):
    """
    For a given .fcf file prepare a bayesian conditional probability plot
    between x_key and y_key.

    :param x_key: Parameter of HklFrame which will be placed on x axis
    :type x_key: str
    :param y_key: Parameter of HklFrame which will be placed on x axis
    :type y_key: str
    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param input_path: Path to the input .fcf file.
    :type input_path: str
    :param input_format: Format of the input .fcf file. For reference see
        :meth:`hikari.dataframes.HklFrame.interpret_hkl_format`.
    :type input_format: int or str or dict
    :param input_wavelength: Wavelength of radiation utilised in experiment.
    :type input_wavelength: float or str
    :param output_path: Path to the output .png file.
    :type output_path: str
    """
    no_of_bins = 10
    p = HklFrame()
    p.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    p.la = input_wavelength
    p.read(make_abspath(input_path), input_format)
    p.place()
    p.calculate_fcf_statistics()
    x = p.table.loc[:, x_key].rank(pct=True).to_numpy()
    y = p.table.loc[:, y_key].rank(pct=True).to_numpy()
    bins = np.zeros(shape=(no_of_bins, no_of_bins))
    lims = [-1.e-8] + [(i + 1) / no_of_bins for i in range(no_of_bins)]
    for i in range(no_of_bins):
        for j in range(no_of_bins):
            bins[i, j] = ((lims[i] < x) & (x <= lims[i+1]) &
                          (lims[j] < y) & (y <= lims[j+1])).sum()
    n_avg = len(x) / no_of_bins ** 2
    chi2 = np.sum((bins - n_avg) ** 2 / n_avg)
    fig = pyplot.figure()
    ax = fig.add_subplot(111, aspect='equal')
    pyplot.xlim(0, 1)
    pyplot.ylim(0, 1)
    h = ax.hist2d(x, y, bins=no_of_bins, alpha=0.25, cmap=cm.get_cmap('PiYG'))
    cb = pyplot.colorbar(h[3], ax=ax)
    cb.set_label('Number of observations')
    ax.scatter(x=x, y=y, s=5.0, c='#000080', marker='.', alpha=0.75)
    pyplot.title('Bayesian CoNditional probability, chi2 = {:.2f}'.format(chi2))
    pyplot.xlabel('"' + x_key + '" rank')
    pyplot.ylabel('"' + y_key + '" rank')
    pyplot.tight_layout()
    pyplot.savefig(fname=make_abspath(output_path), dpi=300)


def observed_vs_calculated_plot(input_path='shelx.fcf',
                                input_format='shelx_fcf',
                                output_path='Io_vs_Ic.png'):
    p = HklFrame()
    p.read(make_abspath(input_path), input_format)
    icalc = p.table.loc[:, 'Ic'].to_numpy()
    iobs = p.table.loc[:, 'I'].to_numpy()
    i_min = min(np.min(icalc[icalc > 0]), np.min(iobs[iobs > 0]))
    i_max = max(np.max(icalc[icalc > 0]), np.max(iobs[iobs > 0]))
    fig = pyplot.figure()
    ax = fig.add_subplot(111)  # , aspect='equal'
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim([i_min, i_max])
    ax.set_ylim([i_min, i_max])
    ax.plot(np.linspace(0, i_max), np.linspace(0, i_max), '-k', lw=1, zorder=0)
    ax.scatter(x=icalc, y=iobs, s=5.0, c='r', marker='.', alpha=0.75, zorder=10)
    pyplot.title('Calculated vs observed intensities plot')
    pyplot.xlabel('I_cal')
    pyplot.ylabel('I_obs')
    pyplot.tight_layout()
    pyplot.savefig(fname=make_abspath(output_path), dpi=300)


def normal_probability_plot(input_path='shelx.fcf',
                            input_format='shelx_fcf',
                            output_path='Io_vs_Ic.png'):

    # scale factors
    a = 0.1000
    b = 0.0

    p = HklFrame()
    p.read(make_abspath(input_path), input_format)
    i_obs = p.table.loc[:, 'I'].to_numpy()
    i_calc = p.table.loc[:, 'Ic'].to_numpy()
    si = p.table.loc[:, 'si'].to_numpy()
    p = 1/3 * i_obs + 2/3 * i_calc
    si = np.sqrt(si ** 2 + (a * p) ** 2 + b * p)

    # expected delta m
    def delta_m(f1, f2, k, si1, si2):
        return np.sort((f1 - k * f2) / np.sqrt(si1 ** 2 + k **2 * si2 ** 2))

    def sum_of_delta_m_squared(k):
        return np.sum(delta_m(i_obs, i_calc, k, si, np.zeros_like(si)) ** 2)

    def scale_factor():
        return minimize(sum_of_delta_m_squared, x0=np.array([1.0])).x[0]

    experiment_delta_m = delta_m(f1=i_obs, f2=i_calc, k=scale_factor(),
                               si1=si, si2=np.zeros_like(si))
    experiment_delta_m = experiment_delta_m / np.std(experiment_delta_m)

    # simulated delta m
    uniform = (np.arange(len(experiment_delta_m))+0.5) / len(experiment_delta_m)
    simulated_delta_m = [erfinv(-1 + 2 * q) for q in uniform]

    # drawing the plot
    fig = pyplot.figure()
    ax = fig.add_subplot(111, aspect='equal')
    ax.set_xlim([-5, 5])
    ax.set_ylim([-5, 5])
    pyplot.hist(experiment_delta_m, bins=100, density=True)
    ax.scatter(experiment_delta_m, simulated_delta_m, s=5.0, c='r', marker='.',
               alpha=0.75, zorder=10)
    ax.plot(np.linspace(-3, 3), np.linspace(-3, 3), '-k', lw=1, zorder=0)
    pyplot.plot(6 * uniform - 3, norm.pdf(6 * uniform - 3))
    pyplot.title('npp')
    pyplot.xlabel('delta_m experiment')
    pyplot.ylabel('delta_m simulated')
    pyplot.tight_layout()
    pyplot.savefig(fname=make_abspath(output_path), dpi=300)


def fcf_descriptors(input_path='shelx.fcf', input_format='shelx_fcf'):
    # scale factors
    a = 0.1000
    b = 0.0

    p = HklFrame()
    p.read(make_abspath(input_path), input_format)
    i_obs = p.table.loc[:, 'I'].to_numpy()
    i_calc = p.table.loc[:, 'Ic'].to_numpy()
    si = p.table.loc[:, 'si'].to_numpy()
    p = 1/3 * i_obs + 2/3 * i_calc
    si_weighted = np.sqrt(si ** 2 + (a * p) ** 2 + b * p)
    ze = (i_obs - i_calc) / si_weighted
    f_calc = np.sqrt(np.abs(i_calc)) * np.sign(i_calc)
    f_obs = np.sqrt(np.abs(i_obs)) * np.sign(i_obs)
    one_over_sf = (2 * abs(i_obs) ** 0.5) / si

    r1 = np.sum(np.abs(f_obs - f_calc)) / np.sum(np.abs(f_obs))
    wr2 = np.sqrt(
        np.sum(np.abs(si_weighted * np.abs(i_obs - i_calc) ** 2)) /
        np.sum(np.abs(si_weighted * i_obs ** 2)))
    awr2 = np.sqrt(
        (np.mean((i_obs - i_calc) ** 2) / np.mean(si_weighted ** 2)) /
        np.mean((i_obs / si_weighted) ** 2))
    gof_if_alpha_equal_one = np.sqrt(np.mean(ze ** 2))
    agof_if_alpha_equal_one = np.sqrt(
        np.mean((i_obs - i_calc) ** 2) /
        np.mean(si_weighted ** 2))

    print('R1    = {:f}'.format(r1))
    print('wR2   = {:f}'.format(wr2))
    print('awR2  = {:f}'.format(awr2))
    print('GoF*  = {:f}'.format(gof_if_alpha_equal_one))
    print('aGoF* = {:f}'.format(agof_if_alpha_equal_one))


def calculate_sample_form_factors(a, b, c, al, be, ga, space_group, res_path):
    """
    Estimate and print selected IAM XRD form factors for given crystal structure

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param space_group: Short Hermann-Mauguin name or index of space group.
        For details see table in hikari.symmetry.space_groups.
    :type space_group: str or int
    :param res_path: Absolute or relative path to the input .res file.
    :type res_path: str
    :return: None
    :rtype: None
    """
    r = ResFrame()
    r.read(make_abspath(res_path))
    r.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    hkl = np.array([(0, 0, 0), (1, 1, 1), (2, 2, 2), (2, 0, 0), (0, 0, 3),
                (1, 0, 1), (1, 1, 8), (5, 0, 2), (4, 4, 0), (2, 0, 6),
                (2, 0, 1), (2, 0, 2), (2, 0, 3), (2, 0, 4), (2, 0, 5),
                (5, 9, 9), (0, 0, 10), (0, 2, 10), (0, 4, 10)])
    f = r.form_factor(np.array(hkl), SG[space_group])
    f2 = f * np.conj(f)
    for _hkl, _f, _f2 in zip(hkl, f, f2):
        print(f'{_hkl}: {_f2:12f} --- {_f}')


if __name__ == '__main__':
    # calculate_sample_form_factors(a=5.64109, b=5.64109, c=5.64109,
    #                               al=90, be=90, ga=90, space_group='Fm-3m',
    #                               res_path='~/x/NaCl/cifmaking/NaCl_more_res.res')
    calculate_sample_form_factors(a=7.210241, b=16.487567, c=11.279203,
                                  al=90, be=90, ga=90, space_group='Pnma',
                                  res_path='~/x/HP/2oAP/_/_.res')



"""hikari - software for handling and analysing crystallographic files.
Copyright (c) 2021, Daniel Tchoń.
All rights reserved.

This source code is licensed under the MIT-style license found in the
LICENSE file in the root directory of this source tree."""

MEMORY_SIZE = 500 * 2 ** 20
"""Approximate upper limit of taken memory in bytes"""



"""
This file contains tools to work with tuples and lists used in the package.
"""

from typing import Sequence

from .typing import Number

import numpy as np


def find_best(strings: list[str], criteria: str) -> str:
    """
    Parse a list of `strings` and return the "best" element based on `criteria`.

    :param strings: List of string where best will be found based on `criteria`.
    :type strings: list[str]
    :param criteria: '>'-separated substrings sought in descending order.
        '+' is a logical 'and', '=' substitutes: `A=B` returns B if A is found.
    :type criteria: str
    :return: Best string based on given `criteria` or `` if no best found.
    :rtype: str
    """
    criterion, _, further_criteria = criteria.partition('>')
    wanted = criterion.partition('=')[0].split('+')
    if all(w in strings or w == '' for w in wanted):
        return criterion.rpartition('=')[2]
    else:
        return find_best(strings, further_criteria)


def cubespace(
        start: Number,
        stop: Number = False,
        num: int = 10,
        include_start: bool = True,
):
    """
    Return sequence of *num* floats between *start* and *stop*.
    Analogously to numpy's linspace, values in returned list are chosen
    so that their cubes (hence name) are spread evenly in equal distance.

    If the parameter *stop* is not given, the value of *start* is used as
    the upper limit instead. In such case the lower limit is set to 0.

    The values of lower limit, *start*, and upper limit, *stop*,
    are included in the list. The *start* value can be excluded
    by setting the *include_start* keyword to False.

    :example:

    >>> cubespace(10, num=3)
    array([ 0.        ,  7.93700526,  10.       ])
    >>> cubespace(0, -10, num=3)
    array([ 0.        , -7.93700526, -10.       ])
    >>> cubespace(0, 10, num=3, include_start=False)
    array([ 6.93361274,  8.73580465, 10.        ])

    :param start: The starting value of a sequence.
    :type start: float
    :param stop: The ending value of a sequence. If False (default), *start*
        is used as the ending value, while the starting value is set to 0.
    :type stop: float
    :param num: Number of samples to generate. Default is 10.
    :type num: int
    :param include_start: If False, the value of *start* is not included in
        returned list. Nonetheless, it is still considered as a starting point.
    :type include_start: bool
    :return: An array with *num* spaced samples in the *start*-*stop* interval.
    :rtype: numpy.ndarray
    """
    (start, stop) = (0.0, start) if stop is False else (start, stop)
    if include_start is False:
        return cubespace(start, stop, num=num+1, include_start=True)[1:]
    cubed_start = pow(start, 3)
    cubed_stop = pow(stop, 3)
    cubed_space = np.linspace(cubed_start, cubed_stop, num)
    return np.sign(cubed_space) * np.power(abs(cubed_space), 1/3)


def rescale_list_to_range(original: Sequence, limits: Sequence) -> list:
    """
    Linearly rescale values in original list to limits (minimum and maximum).

    :example:

    >>> rescale_list_to_range([1, 2, 3], (0, 10))
    [0.0, 5.0, 10.0]
    >>> rescale_list_to_range([1, 2, 3], (-10, 0))
    [-10.0, -5.0, 0.0]
    >>> rescale_list_to_range([1, 2, 3], (0j, 10j))
    [0j, 5j, 10j]

    :param original: Original list or list-like to be rescaled.
    :type original: list
    :param limits: Sequence of two floats, min and max, to constrain the new list
    :type limits: Sequence
    :return: Original list rescaled to fit between min and max
    :rtype: list
    """
    original = np.array(original)
    new_min, new_max = limits[0:2]
    old_min, old_max = min(original), max(original)
    return (new_max + new_min) / 2 * original / old_min if old_min == old_max \
        else list((new_max * (original - old_min) / (old_max - old_min) +
                   new_min * (old_max - original) / (old_max - old_min)))


def rescale_list_to_other(source: Sequence, target: Sequence) -> list:
    """
    Linearly rescale *source* list of numeral values to
    elements of iterable scale in *target*.
    The numeric values in the first list are
    rescaled to the length of *other* using :func:`rescale_list_to_range`,
    changed to integers and used as pointers in *other* to retrieve final value.

    :example:

    >>> rescale_list_to_other([1, 2, 3], [-7.7, -6.6, -5.5, -4.4, -3.3])
    [-7.7, -5.5, -3.3]
    >>> rescale_list_to_other([-7.7, -6.6, -5.5, -4.4, -3.3], [1, 2, 3])
    [1, 1, 2, 3, 3]
    >>> rescale_list_to_other([1, 2, 3], 'holy grail')
    ['h', ' ', 'l']

    :param source: Sequence of numerals to be rescaled to *other*.
    :type source: Sequence
    :param target: Sequence from which the values in new list will be selected.
    :type target: Sequence
    :return: List with elements from *other* assigned using values in original.
    :rtype: list
    """
    return [target[int(v)] for v in
            rescale_list_to_range(original=source, limits=(0, len(target)-1))]



"""
This module contains all dataframes utilised in hikari.
A dataframe is a low-level object, which stores and manipulates certain
crystallographic information. At the moment, the following dataframes
are implemented:

- **hkl** - for single crystal reflection data
- **cif** - for crystallographic open format data (partially)
- **res** - for shelx crystal structure data (partially)

Please mind that the `hkl` frame, HklFrame is the most developed.
Other frames are in an early stage of development.
"""

from .cif import CifFrame, CifBlock
from .base import BaseFrame
from .ubase import UBaseFrame
from .hkl import HklFrame
from .res import ResFrame
from .lst import LstFrame



import abc
from pathlib import Path
import matplotlib as mpl
from matplotlib import pyplot, colors, cm
from mpl_toolkits.mplot3d import art3d
import numpy as np

from hikari.utility import gnuplot_map_palette, make_abspath, \
    mpl_map_palette, sph2cart
from hikari.resources import gnuplot_angular_heatmap_template


class ArtistError(Exception):
    """Exception raised when a problem with plotting in hikari occurs."""

    def __init__(self, message):
        super().__init__(message)


class ArtistFactory:
    """A factory method for creating artists."""
    def __init__(self):
        self._artists = {}

    def register(self, name, artist):
        self._artists[name] = artist

    def create(self, name, **kwargs):
        artist = self._artists.get(name)
        if not artist:
            raise ValueError(f'Artist called "{name}" has not been registered!')
        return artist(**kwargs)


class Artist:
    """Base class used for plotting matplotlib and gnuplot plots"""

    @staticmethod
    def _assert_is_iterable(iterable, length=0):
        try:
            _ = len(iterable)
            if not len(iterable) == length and length > 0:
                raise TypeError()
        except TypeError:
            raise ArtistError(f'object {iterable} should be an iterable'
                              f' of length {length}' if length else '')

    @abc.abstractmethod
    def plot(self, path):
        pass


class AngularHeatmapArtist(Artist, abc.ABC):
    """Base class used for plotting angular heatmaps"""

    HEAT_EXTENSION = '.lst'
    HISTOGRAM_EXTENSION = '.his'

    def __init__(self):
        self.histogram = False
        self._x_axis = ()
        self._y_axis = ()
        self._z_axis = ()
        self._focus = ()
        self._heat_limits = ()
        self._polar_limits = ()
        self._azimuth_limits = ()

    @property
    def x_axis(self):
        return self._x_axis

    @x_axis.setter
    def x_axis(self, a):
        self._assert_is_iterable(a, 3)
        self._x_axis = tuple(a)

    @property
    def y_axis(self):
        return self._y_axis

    @y_axis.setter
    def y_axis(self, a):
        self._assert_is_iterable(a, 3)
        self._y_axis = tuple(a)

    @property
    def z_axis(self):
        return self._z_axis

    @z_axis.setter
    def z_axis(self, a):
        self._assert_is_iterable(a, 3)
        self._z_axis = tuple(a)

    @property
    def heat_limits(self):
        return self._heat_limits

    @heat_limits.setter
    def heat_limits(self, lims):
        self._assert_is_iterable(lims, 2)
        self._heat_limits = tuple(lims)

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, coords):
        self._assert_is_iterable(coords)
        [self._assert_is_iterable(c, 3) for c in coords]
        self._focus = tuple(coords)

    @property
    def polar_limits(self):
        return self._polar_limits

    @polar_limits.setter
    def polar_limits(self, lims):
        self._assert_is_iterable(lims, 2)
        self._polar_limits = tuple(lims)

    @property
    def azimuth_limits(self):
        return self._azimuth_limits

    @azimuth_limits.setter
    def azimuth_limits(self, lims):
        self._assert_is_iterable(lims, 2)
        self._azimuth_limits = tuple(lims)


class GnuplotArtist(abc.ABC):
    """Base class used for plotting gnuplot plots"""

    GNUPLOT_EXTENSION = '.gnu'

    def __init__(self):
        super().__init__()
        self._heat_palette = gnuplot_map_palette['']

    @property
    def heat_palette(self):
        return self._heat_palette

    @heat_palette.setter
    def heat_palette(self, name):
        self._heat_palette = gnuplot_map_palette[name]


class MatplotlibArtist(abc.ABC):
    """Base class used for plotting matplotlib plots"""

    def __init__(self):
        super().__init__()
        self._heat_palette = colors.LinearSegmentedColormap.from_list(
            'heatmap', mpl_map_palette[''], N=256)

    @property
    def heat_palette(self):
        return self._heat_palette

    @heat_palette.setter
    def heat_palette(self, name):
        self._heat_palette = colors.LinearSegmentedColormap.from_list(
            'heatmap', mpl_map_palette[name], N=256)


class GnuplotAngularHeatmapArtist(GnuplotArtist, AngularHeatmapArtist):
    """Base class used for plotting gnuplot angular heatmaps"""

    template = gnuplot_angular_heatmap_template

    @property
    def focus_string(self):
        label = "set label at {}, {}, {} '' point ls 10 front"
        return '\n'.join([label.format(*f) for f in self.focus])

    def plot(self, path):
        png_path = Path(make_abspath(path))
        directory, stem, ext = png_path.parent, png_path.stem, png_path.suffix
        gnu_name = png_path.stem + self.GNUPLOT_EXTENSION
        gnu_path = Path().joinpath(directory, gnu_name)
        s = gnuplot_angular_heatmap_template.format(
            axis_x1=self.x_axis[0], axis_x2=self.x_axis[1],
            axis_x3=self.x_axis[2], axis_y1=self.y_axis[0],
            axis_y2=self.y_axis[1], axis_y3=self.y_axis[2],
            axis_z1=self.z_axis[0], axis_z2=self.z_axis[1],
            axis_z3=self.z_axis[2],
            cplt_min=self.heat_limits[0],
            cplt_max=self.heat_limits[1],
            histogram=int(self.histogram),
            focus_string=self.focus_string,
            job_name=stem,
            min_ph=self.azimuth_limits[0],
            max_ph=self.azimuth_limits[1],
            min_th=self.polar_limits[0],
            max_th=self.polar_limits[1],
            palette=self.heat_palette)
        with open(gnu_path, 'w+') as f:
            f.write(s)
        try:
            from os import system, getcwd
            system('cd ' + str(directory) + '; gnuplot ' + gnu_name)
        except OSError:
            raise ArtistError(f'OSError passed: Cannot plot {gnu_name}')


class MatplotlibAngularHeatmapArtist(MatplotlibArtist, AngularHeatmapArtist):
    """Base class used for plotting matplotlib angular heatmaps"""

    MESH_EXTENSION = '.dat'

    def plot(self, path):
        # OS and I/O operations
        png_path = Path(make_abspath(path))
        directory, stem, ext = png_path.parent, png_path.stem, png_path.suffix
        mesh_name = png_path.stem + self.MESH_EXTENSION
        mesh_path = Path().joinpath(directory, mesh_name)
        heat_mesh = np.loadtxt(str(mesh_path))

        # set-up the plot
        fig = pyplot.figure(figsize=(8, 6))
        ax = fig.add_subplot(111, projection='3d')
        pyplot.rcParams.update({'font.size': 16})
        ax.view_init(elev=90 - sum(self.polar_limits) / 2,
                     azim=sum(self.azimuth_limits) / 2)
        mpl_version = [int(i) for i in mpl.__version__.split('.')]
        if mpl_version >= [3, 6]:  # introduced in mpl 3.3.*
            ax.set_box_aspect(None, zoom=1.4)
        else:  # deprecated mpl 3.6.*, privatized 3.8.*; similar, not identical
            ax.dist = 6.5
        ax.plot([1], [1], [1], 'w')
        ax.plot([-1], [-1], [-1], 'w')
        ax.set_axis_off()

        # prepare surface in cartesian coordinates
        polar_range = np.linspace(start=self.polar_limits[0],
                                  stop=self.polar_limits[1],
                                  num=heat_mesh.shape[1])  # 1=angle_res
        azimuth_range = np.linspace(start=self.azimuth_limits[0],
                                    stop=self.azimuth_limits[1],
                                    num=heat_mesh.shape[0])
        polar_mesh, azimuth_mesh = np.meshgrid(polar_range, azimuth_range)
        x_mesh, y_mesh, z_mesh = sph2cart(r=np.ones_like(polar_mesh),
                                          p=np.deg2rad(polar_mesh),
                                          a=np.deg2rad(azimuth_mesh))
        try:
            np.warnings.filterwarnings('ignore',  # mpl uses depreciated numpy
                                       category=np.VisibleDeprecationWarning)
        except AttributeError:
            pass
        ax.plot_wireframe(x_mesh, y_mesh, z_mesh, colors='k', linewidth=0.25)

        # color map declarations
        m = cm.ScalarMappable(cmap=self.heat_palette)
        m.set_array(heat_mesh)
        m.set_clim(*self.heat_limits)
        pyplot.colorbar(m, ax=ax, fraction=0.06, pad=0.0, shrink=0.9)
        norm = colors.Normalize(*self.heat_limits)

        # draw (100), (010), (010) axes
        len_ = 1.25
        xa, xb = self.x_axis, [_ * len_ for _ in self.x_axis]
        ya, yb = self.y_axis, [_ * len_ for _ in self.y_axis]
        za, zb = self.z_axis, [_ * len_ for _ in self.z_axis]
        ax.add_line(art3d.Line3D((xa[0], xb[0]), (xa[1], xb[1]), (xa[2], xb[2]),
                                 color='r', linewidth=5, zorder=9))
        ax.text(*xb, '(100)', zorder=10, color='k')
        ax.add_line(art3d.Line3D((ya[0], yb[0]), (ya[1], yb[1]), (ya[2], yb[2]),
                                 color='g', linewidth=5, zorder=9))
        ax.text(*yb, '(010)', zorder=10, color='k')
        ax.add_line(art3d.Line3D((za[0], zb[0]), (za[1], zb[1]), (za[2], zb[2]),
                                 color='b', linewidth=5, zorder=9))
        ax.text(*zb, '(001)', zorder=10, color='k')

        # draw focus points
        if self.focus:
            xf, yf, zf = np.array(self.focus).T
            ax.plot(xf, yf, zf, linestyle='none', marker='D', markersize=10,
                    markerfacecolor='none', markeredgewidth=2,
                    markeredgecolor='k', zorder=8)

        # prepare smaller heat mesh for polygon centers and plot the heatmap
        face_heat_mesh = (heat_mesh[1:, 1:] + heat_mesh[1:, :-1] +
                          heat_mesh[:-1, 1:] + heat_mesh[:-1, :-1]) / 4
        color_mesh = self.heat_palette(norm(face_heat_mesh))
        for item in [fig, ax]:
            item.patch.set_visible(False)
        ax.plot_surface(x_mesh, y_mesh, z_mesh, rstride=1, cstride=1,
                        cmap=self.heat_palette, linewidth=0,
                        antialiased=False, facecolors=color_mesh)
        pyplot.subplots_adjust(left=0.0, bottom=0.0, right=0.95, top=1.0)
        pyplot.savefig(png_path, dpi=100, format='png', bbox_inches=None)


artist_factory = ArtistFactory()
artist_factory.register(name='gnuplot_angular_heatmap_artist',
                        artist=GnuplotAngularHeatmapArtist)
artist_factory.register(name='matplotlib_angular_heatmap_artist',
                        artist=MatplotlibAngularHeatmapArtist)



import numpy as np
from uncertainties import ufloat, ufloat_fromstr, unumpy
from uncertainties.umath import sin as usin, cos as ucos
from uncertainties.umath import acos as uacos, sqrt as usqrt

from hikari.dataframes import BaseFrame


class UBaseFrame(BaseFrame):
    """
    A sub-class of :class:`hikari.dataframes.BaseFrame` capable of the same
    operation as its parent, but using `uncertainty.ufloats` instead of floats.
    As a result, types specified in docstring might be wrong due to inheritance.
    """

    IMPORTED_FROM_CIF = {
        'a': ['_cell_length_a', ufloat_fromstr, 1.0],
        'b': ['_cell_length_b', ufloat_fromstr, 1.0],
        'c': ['_cell_length_c', ufloat_fromstr, 1.0],
        'al': ['_cell_length_alpha', ufloat_fromstr, 90],
        'be': ['_cell_length_beta', ufloat_fromstr, 90],
        'ga': ['_cell_length_gamma', ufloat_fromstr, 90],
        'ub11': ['_diffrn_orient_matrix_UB_11', float, 1.0],
        'ub12': ['_diffrn_orient_matrix_UB_12', float, 0.0],
        'ub13': ['_diffrn_orient_matrix_UB_13', float, 0.0],
        'ub21': ['_diffrn_orient_matrix_UB_21', float, 0.0],
        'ub22': ['_diffrn_orient_matrix_UB_22', float, 1.0],
        'ub23': ['_diffrn_orient_matrix_UB_23', float, 0.0],
        'ub31': ['_diffrn_orient_matrix_UB_31', float, 0.0],
        'ub32': ['_diffrn_orient_matrix_UB_32', float, 0.0],
        'ub33': ['_diffrn_orient_matrix_UB_33', float, 1.0]}

    def __init__(self):
        super(UBaseFrame, self).__init__()
        u0, u1, upi = ufloat(0., 0), ufloat(1., 0), ufloat(np.pi, 0)
        u_eye = unumpy.uarray(np.eye(3), np.zeros([3, 3]))
        self._a_d = self._b_d = self._c_d = u1
        self._a_r = self._b_r = self._c_r = u1
        self._al_d = self._be_d = self._ga_d = upi / 2
        self._al_r = self._be_r = self._ga_r = upi / 2
        self._a_v, self._b_v, self._c_v = u_eye
        self._a_w, self._b_w, self._c_w = u_eye
        self._refresh_cell()
        self.orientation = np.array(((1.0, 0, 0), (0, 1.0, 0), (0, 0, 1.0)))
        """3x3 matrix describing orientation of crystal during experiment."""

    def _refresh_cell(self):
        a, b, c = self.a_d, self.b_d, self.c_d
        sa, sb, sg = usin(self.al_d), usin(self.be_d), usin(self.ga_d)
        ca, cb, cg = ucos(self.al_d), ucos(self.be_d), ucos(self.ga_d)
        v = a * b * c * usqrt(1 - ca**2 - cb**2 - cg**2 + 2 * ca * cb * cg)

        u0 = ufloat(0., 0)
        self._a_v = np.array([a, u0, u0])
        self._b_v = np.array([b * cg, b * sg, u0])
        self._c_v = np.array([c * cb, c * (ca - cb * cg)/sg, v/(a * b * sg)])

        self._a_r = b * c * sa / v
        self._b_r = c * a * sb / v
        self._c_r = a * b * sg / v
        self._al_r = uacos((cb * cg - ca) / (sb * sg))
        self._be_r = uacos((cg * ca - cb) / (sg * sa))
        self._ga_r = uacos((ca * cb - cg) / (sa * sb))

        self._a_w = np.cross(self.b_v, self.c_v) / v
        self._b_w = np.cross(self.c_v, self.a_v) / v
        self._c_w = np.cross(self.a_v, self.b_v) / v



"""
This file contains class and definitions of 3-dimensional symmetry operations.
"""

import numpy as np
from fractions import Fraction
from enum import Enum
from typing import Sequence, Union


class Operation:
    """
    Class storing information about symmetry operations, with clear string
    representation and intuitive syntax for combining / utilising operations:
    "=" to compare two symmetry operations for logical equivalence,
    "*" to combine two symmetry operations or transform a vector,
    "** n" to apply symmetry operation n times,
    "% n" to restrict symmetry operation to n unit cells.
    Some of the functions may work incorrectly for rhombohedral unit cells #TODO
    """
    class Type(Enum):
        """Enumerator class storing information about type of operation"""
        rotoinversion = 4
        identity = 3
        reflection = 2
        rotation = 1
        inversion = 0
        rototranslation = -1
        transflection = -2
        translation = -3

    def __init__(self, transformation, translation=np.array([0, 0, 0])) -> None:
        self.tf = np.array(transformation)
        self.tl = np.array(translation)

    def __eq__(self, other: 'Operation') -> bool:
        if isinstance(other, Operation):
            return (np.array_equal(self.tf, other.tf)
                    and np.array_equal(self._tl24, other._tl24))  # noqa
        return NotImplemented

    def __mul__(self, other: 'Operation') -> 'Operation':
        assert isinstance(other, Operation)
        return self.__class__(self.tf @ other.tf, self.tf @ other.tl + self.tl)

    def __pow__(self, power: int, modulo=None) -> 'Operation':
        return self.__class__.from_matrix(np.linalg.matrix_power(self.matrix, power))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(np.{repr(self.tf)}, np.{repr(self.tl)})'.\
            replace('\n', '').replace(' ', '')

    def __str__(self) -> str:
        origin = ','.join([str(Fraction(o).limit_denominator(12))
                           for o in self.origin])
        return self.name + ': ' + self.code + ' (' + origin + ')'

    def __hash__(self) -> int:
        return hash(repr(self))

    @classmethod
    def from_code(cls, code: str) -> 'Operation':
        """
        Create new symmetry operation using symmetry code "x',y',z'" containing
        transformation of each individual coordinate from x,y,z to x',y',z'.

        :param code: string representing new coordinates after operation
        :return: Symmetry operation generated from given coordinate triplet code
        """
        coords = str(code).replace(';', ',').replace(' ', '').lower().split(',')
        tf = np.zeros((3, 3))
        tl = np.zeros(3)
        for i, coord in enumerate(coords):
            coord = coord if coord[0] in '+-' else '+' + coord
            tf[i, 0] = 1 if '+x' in coord else -1 if '-x' in coord else 0
            tf[i, 1] = 1 if '+y' in coord else -1 if '-y' in coord else 0
            tf[i, 2] = 1 if '+z' in coord else -1 if '-z' in coord else 0
            coord = coord.replace('+x', '').replace('+y', '').replace('+z', '')\
                .replace('-x', '').replace('-y', '').replace('-z', '')
            tl[i] = 0 if coord == '' else float(Fraction(coord))
        return cls(tf, tl)

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> 'Operation':
        """
        Create new symmetry operation using augmented 4x4 transformation matrix

        :param matrix: augmented 4x4 matrix
        :return: Symmetry operation generated based on augmented matrix
        """
        return cls(matrix[0:3, 0:3], matrix[0:3, 3])

    @classmethod
    def from_pair(cls, matrix: np.ndarray, vector: np.ndarray) -> 'Operation':
        """
        Create new symmetry operation using point transformation 3x3 matrix
        and 3-length translation vector. Alias for standard creation method.

        :param matrix: 3x3 point transformation matrix
        :param vector: 3-length translation vector
        :return: Symmetry operation generated based on matrix - vector pair
        """
        return cls(matrix, vector)

    @staticmethod
    def _row_to_str(xyz: np.ndarray[int], r: float) -> str:
        """
        Convert xyz: 3-el. list and r - number to single element of code triplet

        :param xyz: 3-element list of coordinates, e.g.: [1,-1,0]
        :param r: translation applied to the row
        :return: string representing change of coordinate
        """
        s = ''
        for var, char in zip(xyz, 'xyz'):
            if np.isclose(abs(var), 0):
                continue
            s += '{:+f}'.format(var)[0]
            if np.isclose(abs(var), 1):
                s += char
                continue
            s += str(Fraction(abs(var)).limit_denominator(9)) + char
        if not np.isclose(abs(r), 0):
            s += '{:+f}'.format(float(r))[0]
            s += str(Fraction(abs(r)).limit_denominator(9))
        return s.strip('+')

    @staticmethod
    def _project(vector: np.ndarray, onto: np.ndarray) -> np.ndarray:
        """Return projection of np.ndarray "vector" to np.ndarray "onto" """
        return (np.dot(vector, onto) / np.sqrt(sum(onto ** 2)) ** 2) * onto

    @property
    def code(self) -> str:
        return ','.join([self._row_to_str(xyz, r) for xyz, r
                         in zip(self.tf, self.tl)])

    @property
    def tf(self) -> np.ndarray:
        return self._tf

    @tf.setter
    def tf(self, value: np.ndarray) -> None:
        self._tf = np.rint(value).astype(int)

    @property
    def tl(self) -> np.ndarray:
        return self._tl24 / 24

    @tl.setter
    def tl(self, value: np.ndarray) -> None:
        self._tl24 = np.rint(value * 24).astype(int)

    @property
    def matrix(self) -> np.ndarray:
        """Augmented 4 x 4 transformation matrix with float-type values"""
        matrix = np.eye(4, dtype=float)
        matrix[0:3, 0:3] = self.tf
        matrix[0:3, 3] = self.tl
        return matrix

    @property
    def det(self) -> int:
        """Determinant of 3x3 transformation part of operation's matrix"""
        return int(np.linalg.det(self.tf))

    @property
    def typ(self) -> Type:
        """Crystallographic type of this symmetry operation (see `Type`)"""
        _trans = self.translational
        if self.trace == 3:
            return self.Type.translation if _trans else self.Type.identity
        elif self.trace == -3:
            return self.Type.inversion
        elif len(self.invariants) == 0:
            return self.Type.rotoinversion
        elif self.det < 0:
            return self.Type.transflection if _trans else self.Type.reflection
        else:
            return self.Type.rototranslation if _trans else self.Type.rotation

    @property
    def name(self) -> str:
        """Short name of symmetry operation, e.g.: 'm', '3' or '2_1'"""
        _glide = self.glide
        _glide_dir = 'n' if np.linalg.norm(_glide) < 1e-8 else 'x'
        d = {(1, 0, 0): 'a', (0, 1, 0): 'b', (0, 0, 1): 'c',
             (0, 1, 1): 'A', (1, 0, 1): 'B', (1, 1, 0): 'C', (1, 1, 1): 'I',
             (1, 2, 2): 'R', (2, 1, 1): 'R', (2, 1, 2): 'R', (1, 2, 1): 'R'}
        for key, value in d.items():
            if np.allclose(_glide, self._project(_glide, np.array(key))):
                _glide_dir = value

        if self.typ is self.Type.identity:
            return '1'
        elif self.typ is self.Type.reflection:
            return 'm'
        elif self.typ is self.Type.rotation:
            return str(self.fold) + (self.sense if self.fold > 2 else '')
        elif self.typ is self.Type.inversion:
            return '-1'
        elif self.typ is self.Type.rototranslation:
            return str(self.fold) + \
                   str(round(np.linalg.norm(_glide)*self.fold)) + self.sense
        elif self.typ is self.Type.transflection:
            return _glide_dir.replace('A', 'n').replace('B', 'n').\
                replace('C', 'n') if self.glide_fold == 2 else 'd'
        elif self.typ is self.Type.translation:
            return _glide_dir
        elif self.typ is self.Type.rotoinversion:
            return str(-self.fold) + self.sense
        else:
            return '?'

    @property
    def fold(self) -> int:
        """
        Number of times operation must be repeated to become identity, inversion
        or translation: n for n-fold axes, 2 for reflections, 1 for other (max 6)
        """
        op = BoundedOperation(self.tf if self.det > 0 else -self.tf)
        for f in (1, 2, 3, 4, 5, 6):
            if (op ** f).trace == 3:
                return f
        raise NotImplementedError('fold is not in range 1 to 6')

    @property
    def order(self) -> int:
        """
        Number of times operation has to be repeated to become
        a translation, e.g.: n for all n-fold axes, 2 for other (max 6)
        """
        op = self.bounded
        for f in (1, 2, 3, 4, 5, 6):
            if pow(op, f, 1).typ is self.Type.identity:
                return f
        raise NotImplementedError('order is not in range 1 to 6')

    @property
    def glide(self) -> np.ndarray:
        """Part of the translation vector stemming from operations' glide"""
        return (self.unbounded ** 24)._tl24 / 576

    @property
    def glide_fold(self) -> int:
        """
        Number of types glide component of the operation must be repeated
        in order to contain only integer values, e.g.: 3 for "6_2", 4 for "d"
        """
        return max([24 // t for t in [*self._tl24, 24] if t != 0])

    @property
    def origin(self) -> np.ndarray:
        """
        Selected point that remains on the symmetry element after operation
        Loosely based on solving https://math.stackexchange.com/q/1054481/.
        """
        return np.linalg.pinv(np.eye(3) - self.tf) @ self.tl

    # TODO Likely incorrect, see ITC 5.2.1. Transformations and example for P4/n
    # TODO Fixed? Test rigorously.

    @property
    def reciprocal(self) -> 'PointOperation':
        """Relevant symmetry operation in its respective reciprocal space"""
        return PointOperation(np.linalg.inv(self.tf).T)

    @property
    def trace(self) -> int:
        """Trace of 3x3 transformation part of operation's matrix"""
        return int(np.trace(self.tf))

    @property
    def translational(self) -> bool:
        """True if operation has any glide component, False otherwise"""
        return not np.allclose(self.glide, 0)

    @property
    def invariants(self) -> list[np.ndarray]:
        """List of directions not affected by this symmetry operation"""
        eigenvalues, eigenvectors = np.linalg.eig(self.tf)
        return eigenvectors.T[np.isclose(eigenvalues.real, 1)].real

    @property
    def orientation(self) -> Union[np.ndarray, None]:
        """Direction of symmetry element (if it can be defined) else None"""
        if self.det < 0:
            o = Operation(-1 * self.tf).orientation
        elif self.typ in {self.Type.rotation, self.Type.rototranslation}:
            m = self.tf  # method: see https://math.stackexchange.com/q/3441262
            t = m + m.T - (m.trace() - 1) * np.eye(3)
            o = [_ for _ in t if not np.isclose(sum(abs(_)), 0)][0]
        else:
            return None
        if o is None:
            return o
        o = o if sum(o) >= 0 else -o
        if self.typ is self.Type.rotoinversion:  # same direction as glide
            o = o if np.dot(o, self.glide) > 0 else -o
        return o / np.sqrt(sum(o*o))

    @property
    def sense(self) -> str:
        """"+" or "-", the "sense" of rotation, as given in ITC A, 11.1.2"""
        unique = np.array([np.sqrt(2), np.e, np.pi])
        rotation = self.tf if self.det > 0 else -self.tf
        sign = np.dot(self.orientation, np.cross(unique, rotation @ unique))
        return '' if np.isclose(sign, 0) else '+' if sign > 0 else '-'

    @property
    def _hm_span(self):
        span = np.array(['', '', ''], dtype='U16')
        for inv in self.invariants:
            coefficients = (inv / min(i for i in abs(inv) if i > 0.01)).astype(int)
            axis = 'xyz'[(coefficients != 0).argmax(axis=0)]
            letters = [str(s).replace('0', '').replace('1', '')
                       + (axis if s != 0 else '') for s in coefficients]
            span += np.array(letters)
        for i, s in enumerate(span):
            if s == '':
                span[i] = Fraction(self.origin[i]).limit_denominator(12)
        return ','.join(f'{s!s:>4s}' for s in span)

    @property
    def _hm_glide(self):
        return ','.join([f'{Fraction(g).limit_denominator(12)!s:>4s}'
                         for g in self.glide])

    @property
    def hm_symbol(self) -> str:
        if self.typ is self.Type.identity:
            return ' 1                                 '
        elif self.typ is self.Type.inversion:
            return f'-1                   {self._hm_span} '
        elif self.typ is self.Type.translation:
            return f' t  ({self._hm_glide})'
        elif self.typ is self.Type.rotation:
            return f' {self.name:2s}                  {self._hm_span}'
        elif self.typ is self.Type.rototranslation:
            return f'{self.name:3s} ({self._hm_glide}) {self._hm_span}'
        elif self.typ is self.Type.reflection:
            return f' m                   {self._hm_span}'
        elif self.typ is self.Type.transflection:
            return f'{self.name:3s} ({self._hm_glide}) {self._hm_span}'
        elif self.typ is self.Type.rotoinversion:
            return f'{self.name:3s} ({self._hm_glide}) {self._hm_span}'
        else:
            return '?'

    @property
    def bounded(self) -> 'BoundedOperation':
        """Instance of self that always collapses down to Coset representative"""
        return self if isinstance(self, BoundedOperation) \
            else BoundedOperation(self.tf, self.tl)

    @property
    def unbounded(self) -> 'Operation':
        """Instance of self that can express translations beyond the unit cube"""
        return Operation(self.tf, self.tl)

    @property
    def is_bounded(self) -> bool:
        """True if self is a coset representatives i.e. 0 <= tl < 1 element-wise"""
        return all(self._tl24 >= 0) and all(self._tl24 <= 24)

    def at(self, point: np.ndarray) -> 'Operation':
        """
        Transform operation as little as possible so that its symmetry element
        contains "point". To be used after "into" if used together.
        Based on "ITC 5.2.1. Transformations".

        :param point: Target coordinates of point which should lie in element
        :return: New symmetry operation which contains "point" in its element
        """
        shift = np.array(point) - self.origin
        for invariant in self.invariants:
            shift -= self._project(shift, onto=invariant)
        return Operation(np.eye(3), shift) * self * Operation(np.eye(3), -shift)

    def into(self, direction: np.ndarray, hexagonal: bool = False) -> 'Operation':
        """
        Rotate operation so that its orientation changes to "direction", while
        preserving fractional glide. To be used before respective "at" method.
        Will most likely not work for unimplemented rhombohedral unit cells.

        :param direction: Target orientation for element of symmetry operation
        :param hexagonal: True if operation is defined in hexagonal coordinates
        :return: New symmetry operation whose orientation is "direction"
        """
        rebase = np.array(((1, -1/2, 0), (0, np.sqrt(3)/2, 0), (0, 0, 1))) \
            if hexagonal else np.eye(3)
        d = rebase @ np.array(direction)
        d /= np.linalg.norm(d)
        o = rebase @ self.orientation
        o /= np.linalg.norm(o)

        def are_parallel(_v, _w):
            return np.isclose(np.dot(_v, _w), 1)

        if o is None:
            return self
        elif are_parallel(d, o):
            return self
        elif are_parallel(-d, o):
            x, y = np.array((1, 0, 0)), np.array((0, 1, 0))
            temp_vector = y if are_parallel(d, x) or are_parallel(d, -x) else x
            return self.into(temp_vector, hexagonal).into(d, hexagonal)
        else:
            p = np.cross(o, d)
            v = np.array(((0, -p[2], p[1]), (p[2], 0, -p[0]), (-p[1], p[0], 0)))
            rot = np.eye(3) + v + (1 / (1 + np.dot(o, d))) * v @ v
            rot_h = np.linalg.inv(rebase) @ rot @ rebase

            if np.allclose(self.glide, 0):
                new_glide = np.array((0, 0, 0))
            else:
                new_glide = rot_h @ self.glide
                glide_els = [abs(g) for g in new_glide if not np.isclose(g, 0)]
                new_glide = (new_glide / min(glide_els)) / self.glide_fold

            return Operation(rot_h @ self.tf @ np.linalg.inv(rot_h),
                             2 * self.origin + new_glide)

    def transform(self, other: np.ndarray) -> np.ndarray:
        """
        Transform a column containing rows of coordinate points

        :param other: A vertical numpy array of coordinate triplets kept in rows
        :return: Same-shaped array of coordinate triplets transformed by self
        """
        if other.shape[1] == 3:
            return (self.tf @ other.T).T + self.tl
        elif other.shape[1] == 4:
            return self.matrix @ other.T
        raise TypeError('Cannot transform "{}"'.format(other))

    def extincts(self, hkl: np.ndarray) -> np.ndarray:
        """
        Return boolean array with truth whenever reflection should be extinct

        :param hkl: An array containing one hkl or multiple hkls in columns
        :return: array of booleans, where extinct reflections are marked as True
        """
        cond1 = np.isclose(hkl, self.reciprocal.transform(hkl)).all(axis=1)
        cond2 = ~np.isclose(np.dot(hkl, self.glide) % 1, 0)
        return cond1 & cond2

    def distance_to_point(self, point: Sequence[Union[int, float]]) -> float:
        if self.typ in {self.Type.rotation, self.Type.rototranslation, self.Type.rotoinversion}:
            point = np.array(point) - self.origin
            projection = np.dot(point, self.orientation) * self.orientation
            orthogonal_vector = point - projection
            return float(np.linalg.norm(orthogonal_vector))
        elif self.typ in {self.Type.reflection, self.Type.transflection}:
            point = np.array(point) - self.origin
            projection = np.dot(point, self.orientation) * self.orientation
            return float(np.linalg.norm(projection))
        elif self.typ in {self.Type.inversion}:
            return float(np.linalg.norm(np.array(point) - self.origin))
        elif self.typ in {self.Type.identity, self.Type.translation}:
            return -1.0
        return NotImplemented


class BoundedOperation(Operation):
    """
    A subclass of `Operation` where all three elements of the translation
    vector are bounded between 0 (inclusive) and 1 (exclusive).
    This class is suitable for handling coset representatives of space groups,
    since upon binding operations related by unit translation become equivalent.
    """

    @property
    def tl(self) -> np.ndarray:
        return self._tl24 / 24

    @tl.setter
    def tl(self, value: np.ndarray) -> None:
        """
        This setter automatically handles binding the translation component
        of the `Operation` into the [0; 1) range.
        This makes the `Operation` a "bounded" "coset representative".
        With this operation 0 casts onto 0, +/-0.5 onto 0.5, and 1 back onto 0.
        """
        self._tl24 = np.rint(value * 24).astype(int) % 24


class PointOperation(BoundedOperation):
    """A subclass of `BoundedOperation`, asserts translation vector = [0,0,0]"""
    @property
    def tl(self) -> np.ndarray:
        return np.array([0, 0, 0], dtype=float)

    @tl.setter
    def tl(self, value: np.ndarray) -> None:
        if np.any(value):
            raise ValueError('Translation in `PointOperation` must be [0,0,0]')
        self._tl24 = np.array([0, 0, 0], dtype=int)



import abc
from collections import UserDict
import pathlib
import re
import tempfile
from typing import Any, Callable, Sequence, TypeVar

from enum import Enum
from functools import lru_cache
from typing import Union, TextIO

from hikari.resources import cif_core_dict
from hikari.utility import make_abspath


T = TypeVar('T')


class CifBlock(UserDict):
    """
    CifBlock object handles all data inside an individual block of Cif file.
    As a subclass of an `UserDict`, in python3.7+ it is ordered by design.
    Individual Cif items can be accessed or assigned using a dict-like syntax.
    """

    def get_as_type(self, key: str, typ: Callable[[Any], T], default: Any = None) -> T:
        """
        Get value of `self[key]` converted to `typ`. If value is a list,
        convert its contents element-wise.

        :param key: key associated with accessed element
        :param typ: type/function applied to a value or its every element
        :param default: if given, return it on KeyError
        :return: value of `self[key]` or `default` converted to `typ`
        """
        value = self.get(key)
        if value is None:
            value = default
        else:
            if isinstance(value, str):
                value = typ(value)
            elif isinstance(value, list):
                value = list(map(typ, value))
            else:
                raise TypeError(f'Unknown value type of {value}: {type(value)}')
        return value

    def read(self, path: str, block: str) -> None:
        """
        Read the contents of .cif file specified by the `path` parameter, but
        access and store only the `block` data block in self.

        :param path: Absolute or relative path to the .cif file.
        :param block: Name of the cif data block to be accessed
        """
        reader = CifReader(cif_file_path=path)
        self.update(reader.read()[block])

    def write(self, path: str) -> None:
        """
        Write the contents of `CifBlock` to the .cif file specified
        by the `path` parameter, using 'hikari' as block name.

        :param path: Absolute or relative path to the .cif file.
        """
        writer = CifWriter(cif_file_path=path)
        writer.write(cif_frame=CifFrame({'hikari': self}))


class CifFrame(UserDict):
    """
    A master object which manages cif files. It utilises other `Cif*` classes
    to manage multiple :class:`CifBlock`s with crystallographic information.
    As a subclass of an `UserDict`, in python3.7+ it is ordered by design.
    Individual Cif blocks and items within them can be accessed or assigned
    using a single- or nested- dict-like syntax.

    Similarly to other `Frame`s, `CifFrame` is designed to work in-place,
    meaning it should be first created, and only then accessed using
    methods such as :func:`read` or :func:`write`, but not chain assignments.

    Unlike dict, CifBlock always initiates empty and does not accept
    any parameters at creation.
    """

    def read(self, path: str) -> None:
        """
        Read the contents of .cif file specified by the `path` parameter.
        Store each found block as a {block_name: CifBlock} pair.

        :param path: Absolute or relative path to the .cif file.
        """
        reader = CifReader(cif_file_path=path)
        self.update(reader.read())

    def write(self, path: str) -> None:
        """
        Write the contents of `CifFrame` to the .cif file specified
        by the `path` parameter.

        :param path: Absolute or relative path to the .cif file.
        """
        writer = CifWriter(cif_file_path=path)
        writer.write(cif_frame=self)


class CifValidator(UserDict):
    """
    This object is used to validate individual cif keys when parsing cif files.
    It knows the metadata about each key based on its entry
    in the cif core dictionary v.2.4.5 packaged with the project.
    Since the specification itself is written in cif format,
    it is also read using the same `CifReader` (but without `CifValidator`).

    Upon initialization, `CifValidator` becomes a dictionary whose
    keys are all valid cif keys, according to the cif specification used.
    Individual values are themselves dictionaries that store information
    about key's contents, `_category`, `_type`, whether they are a `_list` etc.

    contains all keys from core cif dictionary. In order
    to access individual values, use `.get()` instead of bracket notation.
    """

    def __init__(self) -> None:
        super().__init__()
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dic_path = str(pathlib.Path(temp_dir) / 'cif_core.dic')
            with open(temp_dic_path, 'w+') as f:
                f.write(cif_core_dict)
            reader = CifReader(cif_file_path=temp_dic_path, validate=False)
            self.update(reader.read())

    def __contains__(self, item) -> bool:
        try:
            _ = self.get(item)
        except KeyError:
            return False
        else:
            return True

    def get(self, key: str, default: UserDict = None) -> UserDict:
        """Get the dictionary containing information about input cif `key`."""
        # def get(self, key: str, default: Any = str) -> UserDict: FAILED
        key, _key = (key[1:], key) if key.startswith('_') else (key, '_' + key)
        value = UserDict()
        try:
            value = self[key]
        except KeyError as e:
            for self_key, self_value in self.items():
                if key.startswith(self_key):
                    if _key in self_value.get('_name', []):
                        value = self_value
            if not value:
                value = default
        return value

    def get__category(self, key: str, default: str = None) -> str:
        """Close equivalent to `self.get(key).get('_category', default)`"""
        value = self.get(key)
        if value is not None:
            _category = value.get('_category', default)
        else:
            _category = default
        return _category

    def get__list(self, key: str, default: bool = None) -> bool:
        """Close equivalent to `self.get(key).get('_list', default) == 'yes'`"""
        value = self.get(key)
        if value is not None:
            got = value.get('_list')
            _list = True if got == 'yes' else False if got == 'no' else None
        else:
            _list = default
        return _list


class CifIOBuffer(abc.ABC):
    """
    An abstract base class for Cif reader and writer buffers.
    Specifies that data can be added and flushed, names and values are stored
    in a list, output is stored in target (dict if reading, file if writing).
    """

    @abc.abstractmethod
    def __init__(self, target: Any) -> None:
        self.names = []
        self.values = []

    @abc.abstractmethod
    def add(self, data):
        pass

    @abc.abstractmethod
    def flush(self):
        pass


class CifIO(abc.ABC):
    """
    A base class for `CifRead` and `CifWrite`. This class and its inheritors
    base on the IUCr File Syntax version 1.1 Working specification available
    [here](`https://www.iucr.org/resources/cif/spec/version1.1/cifsyntax`)
    """
    COMMENT_REGEX = \
        re.compile(r"(?<=\s)(#.*)(?=$)|(?<=^)(#.*)(?=$)", flags=re.M)
    MATCHING_QUOTES_REGEX = re.compile(r"(\B[\"'])((?:\\\1|(?!\1\s).)*.)(\1\B)")
    MATCHING_OUTER_DELIMITERS_REGEX = \
        re.compile(r"(?<=^)([\"';])([\S\s]*)(\1)(?=$)")
    MULTILINE_QUOTE_REGEX = re.compile(r"(^;)([\S\s]*?)(\n;)", flags=re.M)

    WHITESPACE_SUBSTITUTES = {' ': '█', '\t': '▄', '\n': '▀'}

    def __init__(self, cif_file_path, validate=True):
        self.file_contents = ''
        self.file_path = make_abspath(cif_file_path)
        self.file_lines = []
        self.validate = validate


class CifReaderBuffer(CifIOBuffer):
    """Buffer for reading data from cif file into `CifReader`"""

    def __init__(self, target: dict) -> None:
        super().__init__(target=target)
        self.target: dict = target

    def add(self, word: str) -> None:
        """Append the word to names or values based on its first char"""
        if word.startswith('_'):
            if self.values:
                self.flush()
            self.names.append(word)
        else:
            self.values.append(CifReader.revert_delimiters_and_whitespace(word))

    def flush(self) -> None:
        """Update the target dict with names and values stored hitherto"""
        d = UserDict()
        lv = len(self.values)
        ln = len(self.names)
        if lv == ln == 0:
            pass
        elif ln == 0:
            raise IndexError(f'Orphan values found while '
                             f'flushing buffer: {self.values}')
        elif lv % ln == 0:
            d.update({n: self.values[i::ln] for i, n in
                      enumerate(self.names)})
        else:
            raise IndexError(
                f'len(values) == {lv} % len(names) == {ln} mus'
                f't be zero: {self.values} % {self.names}')
        self.target.update(d)
        self.__init__(target=self.target)


class CifReader(CifIO):
    """
    A helper class managing reading cif files into
    :class:`~.CifFrame` or :class:`~.CifBlock`.
    """

    @property
    def blocks(self) -> dict[str, int]:
        """A dict of block names:line numbers where they start in cif file."""
        return self._blocks(lines=tuple(self.file_lines))

    @lru_cache(maxsize=1)
    def _blocks(self, lines: Sequence[str]) -> dict[str, int]:
        return {l[5:]: i for i, l in enumerate(lines) if l.startswith('data_')}

    class State(Enum):
        """This class stores current cif reading state (e.g. inside loop etc.)"""
        default = 0
        loop_keys = 1
        loop_values = 2

    def format_dictionary(self, parsed_dict_: dict[str, list[str]]) \
            -> dict[str, Union[str, list[str]]]:
        """
        Reformat a dictionary of parsed data so that the format of every name
        and value agrees with the cif core dictionary stored in `CifValidator`.

        :param parsed_dict_: Dictionary with data pairs
        :return: Data dictionary with correctly formatted data names and values
        """

        def item_value_should_be_a_list(k_, v_):
            is_listable = cif_core_validator.get__list(k_) \
                if self.validate else False
            is_long = len(v_) > 1
            is_a_validator_name_field = not self.validate and k_ == '_name'
            return is_listable or is_long or is_a_validator_name_field

        new_dict = dict()
        for k, v in parsed_dict_.items():
            if item_value_should_be_a_list(k, v):
                new_dict[k] = v
            else:
                new_dict[k] = v[0]
        return new_dict

    def parse_lines(self, start: int, end: int) -> dict:
        """
        Read the data from :attr:`~.CifIO.lines` numbered `start` to `end`,
        interpret it, and return it as an instance of a dict.

        :param start: number of the first line which data should be read from
        :param end: number of the first line which should not be read anymore
        :return: ordered dictionary with name: value pairs for all parsed lines
        """
        parsed_data = dict()
        buffer = CifReaderBuffer(target=parsed_data)
        state = self.State.default
        for line in self.file_lines[start:end]:
            if line.lstrip().startswith('loop_'):
                buffer.flush()
                state = self.State.loop_keys
                line = line.lstrip()[5:]
            words = line.strip().split()
            if not words:
                if state is self.State.loop_values:
                    state = self.State.default
                continue
            if words[0].startswith('_') and state is not self.State.loop_keys:
                buffer.flush()
            if not words[0].startswith('_') and state is self.State.loop_keys:
                state = self.State.loop_values
            for word in words:
                buffer.add(word)
        buffer.flush()
        formatted_data = self.format_dictionary(parsed_data)
        return formatted_data

    def read(self) -> dict:
        """
        Read the contents of cif currently pointed by :attr:`~.CifIO.file_path`
        and block :attr:`~.CifIO.data_block_header` and return them to a dict.

        :return: A dictionary containing information read from .cif file.
        """
        with open(self.file_path, 'r') as cif_file:
            self.file_contents = cif_file.read()
        self.protect_multilines()
        self.protect_quotes()
        self.remove_comments()
        self.file_lines = self.file_contents.splitlines()
        block_names = self.blocks.keys()
        block_starts = self.blocks.values()
        block_ends = list(block_starts)[1:] + [None]
        read_data = {}
        for n, s, e in zip(block_names, block_starts, block_ends):
            read_data[n] = CifBlock(self.parse_lines(s + 1, e))
        return read_data

    def protect_multilines(self) -> None:
        """
        Replace whitespace between every pair of "\\n;" sequences with
        substitutes and remove the outer semicolons in `self.file_contents`.
        """

        split_string = self.MULTILINE_QUOTE_REGEX.split(self.file_contents)
        self.file_contents = self._protect_split(split_string)

    def protect_quotes(self) -> None:
        """
        Replace whitespace between every pair of matching quotation marks
        (single or double) with substitutes and remove the outer quotation
        marks in `self.contents`. See stack overflow /q/46967465/ for details.
        """
        split_string = self.MATCHING_QUOTES_REGEX.split(self.file_contents)
        self.file_contents = self._protect_split(split_string)

    @classmethod
    def _protect_split(cls, split_string: list[str]) -> str:
        quoted = split_string[2::4]
        for ws, sub in cls.WHITESPACE_SUBSTITUTES.items():
            quoted = [w.replace(ws, sub) for w in quoted]
        right_delimiters = [w.strip('\n') for w in split_string[3::4]]
        split_string[2::4] = quoted
        split_string[3::4] = right_delimiters
        a = ''.join(split_string)
        return ''.join(split_string)

    def remove_comments(self) -> None:
        """
        Replace all comment blocks (whitespace or start-of-file followed by "#")
        within `self.file_contents` with empty strings.
        """
        self.file_contents = self.COMMENT_REGEX.sub('', self.file_contents)

    @classmethod
    def revert_delimiters_and_whitespace(cls, string: str) -> str:
        """
        If present, remove outer delimiters (matching quotes or semicolons) from
        supplied string, remove `self.WHITESPACE_SUBSTITUTES` and return string.

        :return: `string` without outer delimiters nor whitespace substitutes.
        """
        s = ''.join(cls.MATCHING_OUTER_DELIMITERS_REGEX.split(string)[::2])
        for whitespace, substitute in cls.WHITESPACE_SUBSTITUTES.items():
            s = s.replace(substitute, whitespace)
        return s


class CifWriterBuffer(CifIOBuffer):
    """Buffer for writing data from `CifReader` into cif file """

    MAX_NAME_LENGTH = 33
    MAX_LINE_LENGTH = 80
    MIN_STEP_LENGTH = 2
    WHITESPACE = {' ', '\t', '\n'}

    def __init__(self, target: TextIO) -> None:
        super().__init__(target=target)
        self.target: TextIO = target
        self.current__category = ''
        self.current__list = False
        self.current_len = 0

    def add(self, data: tuple) -> None:
        k_, v_ = data
        k__category = cif_core_validator.get__category(k_)
        k__list = cif_core_validator.get__list(k_) or isinstance(v_, list)
        v_len = len(v_) if isinstance(v_, list) else 0

        cat_match = self.current__category == k__category
        lis_match = self.current__list == k__list
        len_match = self.current_len == v_len
        start_match = k_[:12] == ['', *self.names][-1][:12]
        entry_missing = cif_core_validator.get(k_) is None

        if len_match and ((cat_match and lis_match) or
                          (entry_missing and start_match)):
            self.names.append(k_)
            self.values.append(v_)
        else:
            self.flush()
            self.names = [k_]
            self.values = [v_]
            self.current__category = k__category
            self.current__list = k__list
            self.current_len = v_len

    def flush(self) -> None:
        s = '\n'
        if self.current__list is True:
            s += self.format_table()
        else:
            for n_, v_ in zip(self.names, self.values):
                s += self.format_line(n_, v_) + '\n'
        self.target.write(s)
        self.names = []
        self.values = []

    def format_line(self, k, v) -> str:
        name_string = f'{k:<{self.MAX_NAME_LENGTH}}'
        step_string = ' ' * self.MIN_STEP_LENGTH
        value_string = self.enquote(v)
        if len(name_string + step_string + value_string) > self.MAX_LINE_LENGTH:
            step_string = '\n '
            value_string = self.enquote(v, force=True)
        if value_string.startswith(';'):
            step_string = '\n'
        return name_string + step_string + value_string

    def format_table(self) -> str:
        column_widths = [max(map(len, v)) for v in self.values]
        if sum(column_widths) + len(column_widths) >= self.MAX_LINE_LENGTH:
            pass  # TODO: break long loop tables rows into multiple
        formatted_string = 'loop_\n'
        for name in self.names:
            formatted_string += f' {name}\n'
        for value_row in list(map(list, zip(*self.values))):
            enquoted_value_row = map(self.enquote, value_row)
            formatted_string += f' {" ".join(enquoted_value_row)}\n'
        return formatted_string

    def enquote(self, text: str, force: bool = False) -> str:
        if text == '':
            quoted = "''"
        elif any(whitespace in text for whitespace in self.WHITESPACE) or force:
            if '\n' in text:
                quoted = f';{text}\n;'
            elif "'" not in text:
                quoted = f"'{text}'"
            elif '"' not in text:
                quoted = f'"{text}"'
            elif '\n;' not in text:
                quoted = f';{text}\n;'
            else:
                raise ValueError(f'Unable to quote text: {text}')
        else:
            quoted = text
        return quoted


class CifWriter(CifIO):
    """
    A helper class managing writing :class:`~.CifFrame` or :class:`~.CifBlock`
    into cif files
    """

    def write(self, cif_frame: CifFrame) -> None:
        with open(self.file_path, 'w') as cif_file:
            buffer = CifWriterBuffer(target=cif_file)
            first_block = True
            for block_name, block in cif_frame.items():
                if not first_block:
                    cif_file.write('\n\n')
                cif_file.write(f'data_{block_name}')
                for data in block.items():
                    buffer.add(data)
                buffer.flush()
                first_block = False


cif_core_validator = CifValidator()



import uncertainties


def cfloat(string):
    """
    Create "certain float" (`ufloat.n`) from string by converting it first
    to `uncertainties.ufloat` and then taking the nominal value.

    :param string: string to be converted
    :type string: str
    :return: float with no uncertainty
    :rtype: float
    """
    return uncertainties.ufloat_fromstr(representation=string).n



"""
This file contains all constants and functions containing purely chemical
information / methods used in the package.
"""

import re

chemical_elements = \
    ('H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
     'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca',
     'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
     'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
     'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn',
     'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd',
     'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb',
     'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
     'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
     'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm',
     'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds',
     'Rg', 'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og')
"""Tuple containing 2-letter symbols of the first 118 chemical elements."""


def split_atom_label(label):
    """
    Split atomic label used in ref or cif formats into computer-readable parts.
    :param label: Short atomic label such as "C12A" or "Fe1".
    :type label: str
    :return: 3-element tuple with element, number, and suffix, as str.
    :rtype: tuple
    """
    label_regex = re.compile(r'(\d+|\s+)')
    label_fragments = label_regex.split(label)
    element = label_fragments[0]
    number = label_fragments[1]
    suffix = ''.join(label_fragments[2:])
    return element, number, suffix



"""
This module contains all major scripts of hikari.
A script is a high-level hikari function, which utilises
a lot of low-level objects such as hikari :mod:`hikari.dataframe`
or :mod:`hikari.symmetry` in order to provide more sophisticated functionality.

Since HklFrame is the most developed :mod:`hikari.dataframe` of hikari,
majority of existing scripts revolve about modification of the .hkl file.
It should be mentioned that writing a hikari script does require elemental
knowledge of other objects within the package,
but should not be a problem after a short study of code.

Users are cordially invited to propose their own scripts or script ideas.
"""
from .hkl_potency import potency_map, potency_vs_dac_opening_angle, \
    potency_violin_plot, dac_potency_around_axis
from .hkl_completeness import completeness_statistics, dac_statistics, \
    simulate_dac, reformat_hkl
from .compare_adps import animate_similarity_index, calculate_similarity_indices
from .r1_map import r1_map



import re
from typing import List, Match, Union

import numpy as np

from hikari.symmetry.operations import BoundedOperation
from hikari.utility import dict_union


class HallSymbol:
    """Parse, interpret, and convert Hall symbol to `hikari.symmetry.Group`"""

    class HallSymbolException(Exception):
        """Exception raised when group can't be generated from the symbol"""

    REGEX = re.compile(
        r"""(-?)([pabcirstf])_(-?)(\d)([*xyz]?)([12345abcnuvwd]*)"""
        r"""(?:_(-?)(\d)(['"xyz]?)([12345abcnuvwd]*))?"""
        r"""(?:_(-?)(\d)\*?([12345abcnuvwd]*))?"""
        r"""(?:_-?(\d)([12345abcnuvwd]*))?(?:_\((\d)_(\d)_(\d)\))?""")
    r"""
    This is a regex which matches every possible lowercase Hall symbol for
    classical 3D space groups. It matches up to 17 groups in total as follows:

    1. Centrosymmetry: `[-]?`
    2. Lattice symbol: `[pabcirstf]` (required)
    3. Generator 1 inversion component: `[-]?`
    4. Generator 1 rotation fold: `\d` (required)
    5. Generator 1 rotation direction: `[*xyz]?`
    6. Generator 1 translation components: `[12345abcnuvwd]*`
    7. Generator 2 inversion component: `[-]?`
    8. Generator 2 rotation fold: `\d`
    9. Generator 2 rotation direction: `['"xyz]?`
    10. Generator 2 translation components: `[12345abcnuvwd]*`
    11. Generator 3 inversion component: `[-]?`
    12. Generator 3 rotation fold: `\d`
    13. Generator 3 translation components: `[12345abcnuvwd]*`
    14. Generator 4 rotation fold: `\d`
    15. Generator 4 translation components: `[12345abcnuvwd]*`
    16. Origin shift constituent 1 expressed as count of 1/12 shifts: `\d`
    17. Origin shift constituent 2 expressed as count of 1/12 shifts: `\d`
    18. Origin shift constituent 3 expressed as count of 1/12 shifts: `\d`
    """

    DIRECTION_SYMBOLS = ("'", '"', '*')
    TRANSLATION_SYMBOLS = '12345abcnuvwd'
    LATTICE_GENERATORS = {
        'p': (BoundedOperation(np.eye(3)),),
        'a': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (0, 1 / 2, 1 / 2))),
        'b': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (1 / 2, 0, 1 / 2))),
        'c': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (1 / 2, 1 / 2, 0))),
        'i': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (1 / 2, 1 / 2, 1 / 2))),
        'r': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (2 / 3, 1 / 3, 1 / 3)),
              BoundedOperation(np.eye(3), (1 / 3, 2 / 3, 2 / 3))),
        's': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (1 / 3, 1 / 3, 2 / 3)),
              BoundedOperation(np.eye(3), (2 / 3, 2 / 3, 1 / 3))),
        't': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (1 / 3, 2 / 3, 1 / 3)),
              BoundedOperation(np.eye(3), (2 / 3, 1 / 3, 2 / 3))),
        'f': (BoundedOperation(np.eye(3)),
              BoundedOperation(np.eye(3), (0, 1 / 2, 1 / 2)),
              BoundedOperation(np.eye(3), (1 / 2, 0, 1 / 2)),
              BoundedOperation(np.eye(3), (1 / 2, 1 / 2, 0))),
    }
    UNIVERSAL_MATRICES = {
        '1': np.eye(3),
    }
    PRINCIPAL_ROTATIONS = {
        'x': {'1': np.eye(3),
              '2': np.diag([1, -1, -1]),
              '3': np.array([(1, 0, 0), (0, 0, -1), (0, 1, -1)]),
              '4': np.array([(1, 0, 0), (0, 0, -1), (0, 1, 0)]),
              '6': np.array([(1, 0, 0), (0, 1, -1), (0, 1, 0)]),},
        'y': {'1': np.eye(3),
              '2': np.diag([-1, 1, -1]),
              '3': np.array([(-1, 0, 1), (0, 1, 0), (-1, 0, 0)]),
              '4': np.array([(0, 0, 1), (0, 1, 0), (-1, 0, 0)]),
              '6': np.array([(0, 0, 1), (0, 1, 0), (-1, 0, 1)]),},
        'z': {'1': np.eye(3),
              '2': np.diag([-1, -1, 1]),
              '3': np.array([(0, -1, 0), (1, -1, 0), (0, 0, 1)]),
              '4': np.array([(0, -1, 0), (1, 0, 0), (0, 0, 1)]),
              '6': np.array([(1, -1, 0), (1, 0, 0), (0, 0, 1)]),},
        '*': {'3': np.array([(0, 0, 1), (1, 0, 0), (0, 1, 0)]),}
    }
    FACE_X_DIAGONAL_ROTATIONS = {
        "1'": np.eye(3),
        '2"': np.array([(-1, 0, 0), (0, 0, 1), (0, 1, 0)]),
        "2'": np.array([(-1, 0, 0), (0, 0, -1), (0, -1, 0)]),
    }
    FACE_Y_DIAGONAL_ROTATIONS = {
        "1'": np.eye(3),
        '2"': np.array([(0, 0, 1), (0, -1, 0), (1, 0, 0)]),
        "2'": np.array([(0, 0, -1), (0, -1, 0), (-1, 0, 0)]),
    }
    FACE_Z_DIAGONAL_ROTATIONS = {
        "1'": np.eye(3),
        '2"': np.array([(0, 1, 0), (1, 0, 0), (0, 0, -1)]),
        "2'": np.array([(0, -1, 0), (-1, 0, 0), (0, 0, -1)]),
    }
    BODY_DIAGONAL_ROTATIONS = {
        '1': np.eye(3),
        '3': np.array([(0, 0, 1), (1, 0, 0), (0, 1, 0)]),
    }
    STATIC_TRANSLATIONS = {
        'a': np.array([1/2, 0, 0]),
        'b': np.array([0, 1/2, 0]),
        'c': np.array([0, 0, 1/2]),
        'n': np.array([1/2, 1/2, 1/2]),
        'u': np.array([1/4, 0, 0]),
        'v': np.array([0, 1/4, 0]),
        'w': np.array([0, 0, 1/4]),
        'd': np.array([1 / 4, 1 / 4, 1 / 4]),
    }
    DYNAMIC_TRANSLATIONS = {
        '3x': {'1': np.array([1 / 3, 0, 0]), '2': np.array([2 / 3, 0, 0])},
        '3y': {'1': np.array([0, 1 / 3, 0]), '2': np.array([0, 2 / 3, 0])},
        '3z': {'1': np.array([0, 0, 1 / 3]), '2': np.array([0, 0, 2 / 3])},
        '4x': {'1': np.array([1 / 4, 0, 0]), '3': np.array([3 / 4, 0, 0])},
        '4y': {'1': np.array([0, 1 / 4, 0]), '3': np.array([0, 3 / 4, 0])},
        '4z': {'1': np.array([0, 0, 1 / 4]), '3': np.array([0, 0, 3 / 4])},
        '6x': {'1': np.array([1 / 6, 0, 0]), '2': np.array([1 / 3, 0, 0]),
               '4': np.array([2 / 3, 0, 0]), '5': np.array([5 / 6, 0, 0])},
        '6y': {'1': np.array([0, 1 / 6, 0]), '2': np.array([0, 1 / 3, 0]),
               '4': np.array([0, 2 / 3, 0]), '5': np.array([0, 5 / 6, 0])},
        '6z': {'1': np.array([0, 0, 1 / 6]), '2': np.array([0, 0, 1 / 3]),
               '4': np.array([0, 0, 2 / 3]), '5': np.array([0, 0, 5 / 6])},
    }

    def __init__(self, hall_symbol: str) -> None:
        self.symbol = hall_symbol

    @property
    def elements(self) -> Union[None, Match]:
        """Return `self.symbol` elements indexed as in `self.HALL_REGEX` doc"""
        return self.REGEX.match(self.symbol)

    @property
    def symbol(self) -> str:
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: str) -> None:
        self._symbol = symbol.lower().replace(' ', '_')

    @property
    def generators(self) -> List[BoundedOperation]:
        elements: re.Match = self.elements

        # lattice / centering generators
        generators: List[BoundedOperation] = list(self.LATTICE_GENERATORS[elements[2]])
        if elements[1] == '-':
            generators.append(BoundedOperation(-np.eye(3)))

        # generator 1
        gen1_inv, gen1_fold, gen1_dir, gen1_tl = elements.group(3, 4, 5, 6)
        tf1 = -np.eye(3) if gen1_inv else np.eye(3)
        gen1_dir = gen1_dir or 'z'
        tf1 = tf1 @ self.PRINCIPAL_ROTATIONS[gen1_dir][gen1_fold]
        tl1 = np.array([0., 0., 0.])
        gen1_translation_dicts = [self.STATIC_TRANSLATIONS]
        if gen1_fold in '346':
            screw_tls = self.DYNAMIC_TRANSLATIONS.get(gen1_fold + gen1_dir, {})
            gen1_translation_dicts.append(screw_tls)
        gen1_translations = dict_union(*gen1_translation_dicts)
        for tl in gen1_tl:
            tl1 += gen1_translations[tl]
        generators.append(BoundedOperation(tf1, tl1))

        # generator 2
        gen2_inv, gen2_fold, gen2_dir, gen2_tl = elements.group(7, 8, 9, 10)
        if gen2_fold:
            tf2 = -np.eye(3) if gen2_inv else np.eye(3)
            if gen1_fold in '24':
                gen2_dir = gen2_dir or "x"
                tf2 = tf2 @ self.PRINCIPAL_ROTATIONS[gen2_dir][gen2_fold]
            else:  # if gen1_fold in '36'
                gen2_dir = gen2_dir or "'"
                diagonal_rots = self.FACE_X_DIAGONAL_ROTATIONS if gen1_dir == 'x' \
                    else self.FACE_Y_DIAGONAL_ROTATIONS if gen1_dir == 'y' \
                    else self.FACE_Z_DIAGONAL_ROTATIONS
                tf2 = tf2 @ diagonal_rots[gen2_fold + gen2_dir]
            tl2 = np.array([0., 0., 0.])
            for tl in gen2_tl:
                tl2 += self.STATIC_TRANSLATIONS[tl]
            generators.append(BoundedOperation(tf2, tl2))

        # generator 3
        gen3_inv, gen3_fold, gen3_tl = elements.group(11, 12, 13)
        if gen3_fold:
            tf3 = -np.eye(3) if gen3_inv else np.eye(3)
            tf3 = tf3 @ self.BODY_DIAGONAL_ROTATIONS[gen3_fold]
            tl3 = np.array([0., 0., 0.])
            for tl in gen3_tl:
                tl3 += self.STATIC_TRANSLATIONS[tl]
            generators.append(BoundedOperation(tf3, tl3))

        # generator 4
        gen4_fold, gen4_tl = elements.group(14, 15)
        if gen4_fold:
            tl4 = np.array([0., 0., 0.])
            for tl in gen4_tl:
                tl4 += self.STATIC_TRANSLATIONS[tl]
            generators.append(BoundedOperation(-np.eye(3), tl4))

        if any(elements.group(16, 17, 18)):
            shift_ints = [int(i) for i in elements.group(16, 17, 18)]
            shift_vector = np.array(shift_ints, dtype=float) / 12
            shift_op_left = BoundedOperation(np.eye(3), shift_vector)
            shift_op_right = BoundedOperation(np.eye(3), -shift_vector)
            generators = [BoundedOperation.from_matrix(shift_op_left.matrix @ g.matrix
                                                @ shift_op_right.matrix) for g in generators]

        return generators



import numpy as np


class Interval:
    """
    A class handling [A, B] line segments and numpy operations on them.
    The following magic methods have been implemented for `Interval`:

    - creation using `Interval(A, B)` syntax (if `A == B`, represents a point)
    - comparison operations with numbers (True if holds for the whole Interval)
    - arithmetic operations, which act element-wise for left and right edge
    - access methods: `A == Interval(A, B)[0] == Interval(A, B).left`
    - container methods: 5 and 6.2 are `in Interval(5, 7)`, but 7. might not be
    """

    # CREATION METHODS
    def __init__(self, left, right):
        self.left = min(left, right)
        self.right = max(left, right)

    # COMPARISON METHODS
    def __eq__(self, other):
        return self.left == other.left and self.right == other.right

    def __lt__(self, other):
        return self.right < other

    def __gt__(self, other):
        return self.left > other

    def __le__(self, other):
        return self.right <= other

    def __ge__(self, other):
        return self.left >= other

    # UNARY OPERATIONS
    def __pos__(self):
        return self

    def __neg__(self):
        return Interval(-self.right, -self.left)

    # ARITHMETIC METHODS
    def __add__(self, other):
        return Interval(self.left + other, self.right + other)

    def __sub__(self, other):
        return Interval(self.left - other, self.right - other)

    def __mul__(self, other):
        return Interval(self.left * other, self.right * other)

    def __truediv__(self, other):
        return Interval(self.left / other, self.right / other)

    # REPRESENTATION METHODS
    def __str__(self):
        return f'[{self.left}, {self.right}]'

    def __repr__(self):
        return f'Interval({self.left}, {self.right})'

    # CONTAINER METHODS
    def __iter__(self):
        yield self.left
        yield self.right

    def __getitem__(self, key):
        return [self.left, self.right][key]

    def __setitem__(self, key, value):
        new_limits = [self.left, self.right]
        new_limits[key] = value
        self.left, self.right = new_limits

    def __contains__(self, item):
        return self.left <= self._min(item) and self._max(item) <= self.right

    def __len__(self):
        return 2

    @staticmethod
    def _min(item):
        """Recursively returns minimum of args if possible, otherwise args"""
        try:
            return min(item)
        except TypeError:
            return item

    @staticmethod
    def _max(item):
        """Recursively returns maximum of args if possible, otherwise args"""
        try:
            return max(item)
        except TypeError:
            return item

    def arange(self, step=1):
        """
        Return a 1D-list of values from left to right every step
        :param step: spacing between adjacent values, default 1.
        :type step: int or float
        :return: array of values from left to right (including right) every step
        :rtype: np.array
        """
        epsilon = step * 1e-7
        size = 1 + int((self.right - self.left + epsilon) / step)
        return np.array([self.left + step * i for i in range(size)])

    def comb_with(self, *others, step=1):
        """
        Return combinations of self.arange(step) with every other.arange(step)
        :param others: interval or iterable of intervals to comb self with
        :type others: Interval or tuple or list
        :param step: spacing between adjacent values, default 1.
        :type step: int or float
        :return: array of all combinations found in numpy.meshgrid every step
        :rtype: np.array
        """
        arrays = [self.arange(step)] + [other.arange(step) for other in others]
        return np.array(np.meshgrid(*arrays)).reshape(len(arrays), -1)

    def mesh_with(self, *others, step=1):
        """
        Return a numpy.mesh of self.arange(step) with every other.arange(step)
        :param others: interval or iterable of intervals to mesh self with
        :type others: Interval or tuple or list
        :param step: spacing between adjacent values, default 1.
        :type step: int or float
        :return: array of values meshed by numpy.meshgrid every step
        :rtype: np.array
        """
        arrays = [self.arange(step)] + [other.arange(step) for other in others]
        return np.meshgrid(*arrays)



class LstFrame:
    def __init__(self):
        pass

    @staticmethod
    def read_r1(path):
        """Read and return the final value of R1 from lst file"""

        # READ THE FILE
        lines = [line.strip() for line in open(path, 'r') if line.strip()]

        # READ THE LINES, LOOK FOR R1, RETURN THE LAST FOUND
        r1 = ''
        for line in lines:
            if 'R1 =' in line:
                r1 = float(line.split()[2])

        return r1
